import json
import os

from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request

from .._types.credentials_cache import CredentialsCache


def get_user_account_credentials(
    *,
    authorized_user_file: str | os.PathLike[str] | None = None,
    authorized_user_data: dict[str, object] | None = None,
    scopes: list[str],
    cache: CredentialsCache | None = None,
) -> Credentials:
    cached_credentials = _get_cached_credentials(
        cache,
        scopes=scopes,
    )

    if cached_credentials:
        credentials = cached_credentials

    elif authorized_user_data:
        credentials = Credentials.from_authorized_user_info(  # type: ignore[no-untyped-call]
            authorized_user_data,
            scopes=scopes,
        )

    elif authorized_user_file:
        authorized_user_file = os.path.expanduser(
            os.path.expandvars(os.fspath(authorized_user_file))
        )

        if not os.path.exists(authorized_user_file):
            raise ValueError(
                f"Authorized user file does not exist: {authorized_user_file}"
            )

        credentials = Credentials.from_authorized_user_file(  # type: ignore[no-untyped-call]
            authorized_user_file,
            scopes=scopes,
        )

    else:
        raise ValueError("No valid user account credentials found.")

    if credentials.valid:
        if cache and not cached_credentials:
            cache.set(credentials.to_json())  # type: ignore[no-untyped-call]

        return credentials

    if credentials.expired and credentials.refresh_token:
        credentials.refresh(Request())  # type: ignore[no-untyped-call]

        if cache:
            cache.set(credentials.to_json())  # type: ignore[no-untyped-call]

    return credentials


def _get_cached_credentials(
    cache: CredentialsCache | None,
    scopes: list[str],
) -> Credentials | None:
    if not cache:
        return None

    cached_json_str = cache.get()

    if not cached_json_str:
        return None

    cached_credentials = Credentials.from_authorized_user_info(  # type: ignore[no-untyped-call]
        json.loads(cached_json_str),
        scopes=scopes,
    )

    return cached_credentials  # type: ignore[no-any-return]



from google.auth import jwt
from google.auth.transport.requests import Request

from .get_credentials import get_credentials
from .._settings import GoogleAuthSettings


def get_self_signed_jwt(
    settings_key: str | None = None,
    *,
    settings: GoogleAuthSettings | None = None,
    audience: str,
) -> str:
    """
    Get a self-signed JWT
    """
    credentials = get_credentials(settings_key, settings=settings)

    jwt_creds = jwt.Credentials.from_signing_credentials(credentials, audience=audience)  # type: ignore[no-untyped-call]
    # Generate a self-signed JWT. Does not communicate over the network
    jwt_creds.refresh(Request())  # type: ignore[no-untyped-call]

    return jwt_creds.token.decode("utf-8")  # type: ignore[no-any-return]



import google.oauth2.service_account
import pytest

from kiarina.lib.google.auth import settings_manager, get_service_account_credentials


def test_file(load_settings):
    settings = settings_manager.get_settings("service_account_file")
    credentials = get_service_account_credentials(
        service_account_file=settings.service_account_file
    )
    assert isinstance(credentials, google.oauth2.service_account.Credentials)


def test_nonexistent_file():
    with pytest.raises(ValueError, match="Service account file does not exist"):
        get_service_account_credentials(
            service_account_file="/path/to/nonexistent/file.json"
        )


def test_data(load_settings):
    settings = settings_manager.get_settings("service_account_data")
    credentials = get_service_account_credentials(
        service_account_data=settings.get_service_account_data()
    )
    assert isinstance(credentials, google.oauth2.service_account.Credentials)



import os

import pytest
from pydantic_settings_manager import load_user_configs

import kiarina.utils.file as kf


@pytest.fixture(scope="session")
def load_settings():
    env_var = "KIARINA_LIB_GOOGLE_AUTH_TEST_SETTINGS_FILE"

    if env_var not in os.environ:
        pytest.skip(f"Environment variable {env_var} not set, skipping tests.")

    test_settings_file = os.environ[env_var]
    test_settings_file = os.path.expanduser(test_settings_file)

    if not os.path.exists(test_settings_file):
        raise FileNotFoundError(f"Settings file not found: {test_settings_file}")

    user_configs = kf.read_yaml_dict(test_settings_file)

    if not user_configs:
        raise ValueError(f"Settings file is empty or invalid: {test_settings_file}")

    load_user_configs(user_configs)



from google.oauth2.credentials import Credentials
import pytest

from kiarina.lib.google.auth import (
    CredentialsCache,
    settings_manager,
    get_user_account_credentials,
)


def test_file(load_settings):
    settings = settings_manager.get_settings("user_account_file")
    credentials = get_user_account_credentials(
        authorized_user_file=settings.authorized_user_file,
        scopes=settings.scopes,
    )
    assert isinstance(credentials, Credentials)


def test_nonexistent_file():
    with pytest.raises(ValueError, match="Authorized user file does not exist"):
        get_user_account_credentials(
            authorized_user_file="/path/to/nonexistent/file.json",
            scopes=["https://www.googleapis.com/auth/devstorage.read_only"],
        )


def test_data(load_settings):
    settings = settings_manager.get_settings("user_account_data")
    credentials = get_user_account_credentials(
        authorized_user_data=settings.get_authorized_user_data(),
        scopes=settings.scopes,
    )
    assert isinstance(credentials, Credentials)


def test_cache(load_settings):
    set_counter = 0

    class InMemoryCache(CredentialsCache):
        def __init__(self):
            self._cache: str | None = None

        def get(self) -> str | None:
            return self._cache

        def set(self, value: str) -> None:
            self._cache = value

            nonlocal set_counter
            set_counter += 1

    cache = InMemoryCache()

    settings = settings_manager.get_settings("user_account_data")
    credentials = get_user_account_credentials(
        authorized_user_data=settings.get_authorized_user_data(),
        scopes=settings.scopes,
        cache=cache,
    )

    assert isinstance(credentials, Credentials)
    assert credentials.valid is True
    assert cache.get() is not None
    assert set_counter == 1

    credentials2 = get_user_account_credentials(
        authorized_user_data=settings.get_authorized_user_data(),
        scopes=settings.scopes,
        cache=cache,
    )

    assert isinstance(credentials2, Credentials)
    assert credentials2.valid is True
    assert credentials2.token == credentials.token
    assert set_counter == 1  # Cache should be used, so set() not called



import logging
from importlib import import_module
from importlib.metadata import version
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ._helpers.get_credentials import get_credentials
    from ._helpers.get_self_signed_jwt import get_self_signed_jwt
    from ._settings import GoogleAuthSettings, settings_manager
    from ._types.credentials import Credentials
    from ._types.credentials_cache import CredentialsCache
    from ._utils.get_default_credentials import get_default_credentials
    from ._utils.get_service_account_credentials import get_service_account_credentials
    from ._utils.get_user_account_credentials import get_user_account_credentials

__version__ = version("kiarina-lib-google-auth")

__all__ = [
    # ._helpers
    "get_credentials",
    "get_self_signed_jwt",
    # ._settings
    "GoogleAuthSettings",
    "settings_manager",
    # ._types
    "Credentials",
    "CredentialsCache",
    # ._utils
    "get_default_credentials",
    "get_service_account_credentials",
    "get_user_account_credentials",
]

logging.getLogger(__name__).addHandler(logging.NullHandler())


def __getattr__(name: str) -> object:
    if name not in __all__:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")

    module_map = {
        # ._helpers
        "get_credentials": "._helpers.get_credentials",
        "get_self_signed_jwt": "._helpers.get_self_signed_jwt",
        # ._settings
        "GoogleAuthSettings": "._settings",
        "settings_manager": "._settings",
        # ._types
        "Credentials": "._types.credentials",
        "CredentialsCache": "._types.credentials_cache",
        # ._utils
        "get_default_credentials": "._utils.get_default_credentials",
        "get_service_account_credentials": "._utils.get_service_account_credentials",
        "get_user_account_credentials": "._utils.get_user_account_credentials",
    }

    globals()[name] = getattr(import_module(module_map[name], __name__), name)
    return globals()[name]



import os

from google.oauth2.service_account import Credentials


def get_service_account_credentials(
    *,
    service_account_file: str | os.PathLike[str] | None = None,
    service_account_data: dict[str, object] | None = None,
    scopes: list[str] | None = None,
) -> Credentials:
    credentials: Credentials

    if service_account_data:
        credentials = Credentials.from_service_account_info(service_account_data)  # type: ignore[no-untyped-call]

    elif service_account_file:
        service_account_file = os.path.expanduser(
            os.path.expandvars(os.fspath(service_account_file))
        )

        if not os.path.exists(service_account_file):
            raise ValueError(
                f"Service account file does not exist: {service_account_file}"
            )

        credentials = Credentials.from_service_account_file(service_account_file)  # type: ignore[no-untyped-call]

    else:
        raise ValueError("No valid service account credentials found.")

    if scopes:
        credentials = credentials.with_scopes(scopes)

    return credentials



from kiarina.lib.google.auth import get_self_signed_jwt


def test_get_self_signed_jwt(load_settings):
    jwt = get_self_signed_jwt(
        "service_account_file",
        audience="https://blazeworks.jp/",
    )
    assert jwt.count(".") == 2



import json
import os
from typing import Any, Literal

from pydantic import Field, SecretStr, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic_settings_manager import SettingsManager


class GoogleAuthSettings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="KIARINA_LIB_GOOGLE_AUTH_")

    type: Literal["default", "service_account", "user_account", "api_key"] = "default"

    # --------------------------------------------------
    # Fields (common)
    # --------------------------------------------------

    project_id: str | None = None

    impersonate_service_account: str | None = None
    """
    Email address of the service account to impersonate

    The source principal requires the roles/iam.serviceAccountTokenCreator role.
    Note that this required permission is not included in the roles/owner role.
    """

    scopes: list[str] = Field(
        default_factory=lambda: [
            "https://www.googleapis.com/auth/cloud-platform",  # All GCP resources
            "https://www.googleapis.com/auth/drive",  # Google Drive resources
            "https://www.googleapis.com/auth/spreadsheets",  # Google Sheets resources
        ]
    )
    """
    List of scopes to request during the authentication

    Specify the scopes required for impersonation authentication.
    Specify the scopes required for user account authentication.
    """

    # --------------------------------------------------
    # Fields (service_account)
    # --------------------------------------------------

    service_account_email: str | None = None

    service_account_file: str | None = None
    """Path to the service account key file"""

    service_account_data: SecretStr | None = None
    """Service account key data in JSON format"""

    # --------------------------------------------------
    # Fields (user_account)
    # --------------------------------------------------

    user_account_email: str | None = None

    client_secret_file: str | None = None
    """Path to the client secret file"""

    client_secret_data: SecretStr | None = None
    """Client secret data in JSON format"""

    authorized_user_file: str | None = None
    """Path to the authorized user file"""

    authorized_user_data: SecretStr | None = None
    """
    Authorized user data in JSON format

    If expired, retrieve from cache.
    If cache is not available, use for refresh.
    Refreshed credentials will be cached.
    """

    # --------------------------------------------------
    # Fields (api_key)
    # --------------------------------------------------

    api_key: SecretStr | None = None
    """API key for accessing Google APIs"""

    # --------------------------------------------------
    # Validators
    # --------------------------------------------------

    @field_validator(
        "service_account_file",
        "client_secret_file",
        "authorized_user_file",
        mode="before",
    )
    @classmethod
    def expand_user(cls, v: str | None) -> str | None:
        return os.path.expanduser(v) if isinstance(v, str) else v

    # --------------------------------------------------
    # Methods
    # --------------------------------------------------

    def get_service_account_data(self) -> dict[str, Any] | None:
        if not self.service_account_data:
            return None

        return json.loads(self.service_account_data.get_secret_value())  # type: ignore[no-any-return]

    def get_client_secret_data(self) -> dict[str, Any] | None:
        if not self.client_secret_data:
            return None

        return json.loads(self.client_secret_data.get_secret_value())  # type: ignore[no-any-return]

    def get_authorized_user_data(self) -> dict[str, Any] | None:
        if not self.authorized_user_data:
            return None

        return json.loads(self.authorized_user_data.get_secret_value())  # type: ignore[no-any-return]


settings_manager = SettingsManager(GoogleAuthSettings, multi=True)



import os

import google.oauth2.credentials
import google.oauth2.service_account
import pytest

from kiarina.lib.google.auth import settings_manager
from kiarina.lib.google.auth._utils.get_default_credentials import (
    get_default_credentials,
)


@pytest.mark.xfail(
    not os.path.exists(
        os.path.expanduser("~/.config/gcloud/application_default_credentials.json")
    ),
    reason="ADC file not set",
)
def test_adc():
    credentials = get_default_credentials()
    print(f"Obtained credentials of type: {type(credentials)}")
    assert isinstance(credentials, google.oauth2.credentials.Credentials)


def test_service_account(load_settings, monkeypatch):
    if "GOOGLE_APPLICATION_CREDENTIALS" not in os.environ:
        settings = settings_manager.get_settings("service_account_file")
        monkeypatch.setenv(
            "GOOGLE_APPLICATION_CREDENTIALS", settings.service_account_file
        )

    credentials = get_default_credentials()
    print(f"Obtained service account credentials of type: {type(credentials)}")
    assert isinstance(credentials, google.oauth2.service_account.Credentials)
